from __future__ import annotations

__all__ = ["numpy_available", "torch_available"]

from pytest import mark

from arctix.utils.imports import is_numpy_available, is_torch_available

numpy_available = mark.skipif(not is_numpy_available(), reason="Requires NumPy")
torch_available = mark.skipif(not is_torch_available(), reason="Requires PyTorch")



__all__ = [
    "BaseFormatter",
    "BaseSummarizer",
    "Summarizer",
    "is_numpy_available",
    "is_torch_available",
    "set_summarizer_options",
    "summarizer_options",
    "summary",
]

from arctix.formatter import BaseFormatter
from arctix.summarizer import (
    BaseSummarizer,
    Summarizer,
    set_summarizer_options,
    summarizer_options,
    summary,
)
from arctix.utils.imports import is_numpy_available, is_torch_available

# Register NumPy comparators
if is_numpy_available():  # pragma: no cover
    from arctix import _numpy  # noqa: F401

# Register PyTorch comparators
if is_torch_available():  # pragma: no cover
    from arctix import _torch  # noqa: F401



from __future__ import annotations

__all__ = [
    "BaseSummarizer",
    "Summarizer",
    "summarizer_options",
    "summary",
]


from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from contextlib import contextmanager
from typing import Any

from arctix.formatter import (
    BaseFormatter,
    DefaultFormatter,
    MappingFormatter,
    SequenceFormatter,
    SetFormatter,
)
from arctix.utils.format import str_indent, str_mapping


def summary(value: Any, max_depth: int = 1, summarizer: BaseSummarizer | None = None) -> str:
    r"""Summarize the input value in a string.

    Args:
    ----
        value: Specifies the value to summarize.
        max_depth (int, optional): Specifies the maximum depth to
            summarize if the input is nested. Default: ``1``
        summarizer (``BaseSummarizer`` or ``None``): Specifies the
            summarization strategy. If ``None``, the default
            ``Summarizer`` is used. Default: ``None``

    Returns:
    -------
        str: The summary as a string.

    Example usage:

    .. code-block:: pycon

        >>> from arctix import summary
        >>> print(summary(1))
        <class 'int'> 1
        >>> print(summary(["abc", "def"]))
        <class 'list'> (length=2)
          (0): abc
          (1): def
        >>> print(summary([[0, 1, 2], {"key1": "abc", "key2": "def"}]))
        <class 'list'> (length=2)
          (0): [0, 1, 2]
          (1): {'key1': 'abc', 'key2': 'def'}
        >>> print(summary([[0, 1, 2], {"key1": "abc", "key2": "def"}], max_depth=2))
        <class 'list'> (length=2)
          (0): <class 'list'> (length=3)
              (0): 0
              (1): 1
              (2): 2
          (1): <class 'dict'> (length=2)
              (key1): abc
              (key2): def
    """
    summarizer = summarizer or Summarizer()
    return summarizer.summary(value=value, depth=0, max_depth=max_depth)


class BaseSummarizer(ABC):
    r"""Define the base class to implement a summarizer."""

    @abstractmethod
    def summary(
        self,
        value: Any,
        depth: int = 0,
        max_depth: int = 1,
    ) -> str:
        r"""Summarize the input value in a string.

        Args:
        ----
            value: Specifies the value to summarize.
            max_depth (int, optional): Specifies the maximum depth to
                summarize if the input is nested. Default: ``1``
            summarizer (``BaseSummarizer`` or ``None``): Specifies the
                summarization strategy. If ``None``, the default
                ``Summarizer`` is used. Default: ``None``

        Returns:
        -------
            str: The summary as a string.

        Example usage:

        .. code-block:: pycon

            >>> from arctix import Summarizer
            >>> print(Summarizer().summary(1))
            <class 'int'> 1
            >>> print(Summarizer().summary(["abc", "def"]))
            <class 'list'> (length=2)
              (0): abc
              (1): def
            >>> print(Summarizer().summary([[0, 1, 2], {"key1": "abc", "key2": "def"}]))
            <class 'list'> (length=2)
              (0): [0, 1, 2]
              (1): {'key1': 'abc', 'key2': 'def'}
            >>> print(Summarizer().summary([[0, 1, 2], {"key1": "abc", "key2": "def"}], max_depth=2))
            <class 'list'> (length=2)
              (0): <class 'list'> (length=3)
                  (0): 0
                  (1): 1
                  (2): 2
              (1): <class 'dict'> (length=2)
                  (key1): abc
                  (key2): def
        """


class Summarizer(BaseSummarizer):
    """Implement the default summarizer.

    The registry is a class variable, so it is shared with all the
    instances of this class.
    """

    registry: dict[type[object], BaseFormatter] = {
        Mapping: MappingFormatter(),
        Sequence: SequenceFormatter(),
        dict: MappingFormatter(),
        list: SequenceFormatter(),
        object: DefaultFormatter(),
        set: SetFormatter(),
        tuple: SequenceFormatter(),
    }

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_mapping(self.registry))}\n)"

    def summary(
        self,
        value: Any,
        depth: int = 0,
        max_depth: int = 1,
    ) -> str:
        return self.find_formatter(type(value)).format(
            summarizer=self,
            value=value,
            depth=depth,
            max_depth=max_depth,
        )

    @classmethod
    def add_formatter(
        cls, data_type: type[object], formatter: BaseFormatter, exist_ok: bool = False
    ) -> None:
        r"""Add a formatter for a given data type.

        Args:
        ----
            data_type: Specifies the data type for this test.
            formatter (``BaseFormatter``): Specifies the formatter
                to use for the specified type.
            exist_ok (bool, optional): If ``False``, ``RuntimeError``
                is raised if the data type already exists. This
                parameter should be set to ``True`` to overwrite the
                formatter for a type. Default: ``False``.

        Raises:
        ------
            RuntimeError if a formatter is already registered for the
                data type and ``exist_ok=False``.

        Example usage:

        .. code-block:: pycon

            >>> from arctix import Summarizer
            >>> from arctix.formatter import MappingFormatter
            >>> Summarizer.add_formatter(dict, MappingFormatter(), exist_ok=True)
        """
        if data_type in cls.registry and not exist_ok:
            raise RuntimeError(
                f"A formatter ({cls.registry[data_type]}) is already registered for the data "
                f"type {data_type}. Please use `exist_ok=True` if you want to overwrite the "
                "formatter for this type"
            )
        cls.registry[data_type] = formatter

    @classmethod
    def has_formatter(cls, data_type: type[object]) -> bool:
        r"""Indicate if a formatter is registered for the given data
        type.

        Args:
        ----
            data_type: Specifies the data type to check.

        Returns:
        -------
            bool: ``True`` if a formatter is registered,
                otherwise ``False``.

        Example usage:

        .. code-block:: pycon

            >>> from arctix import Summarizer
            >>> Summarizer.has_formatter(list)
            True
            >>> Summarizer.has_formatter(str)
            False
        """
        return data_type in cls.registry

    @classmethod
    def find_formatter(cls, data_type: Any) -> BaseFormatter:
        r"""Find the formatter associated to an object.

        Args:
        ----
            data_type: Specifies the data type to get.

        Returns:
        -------
            ``BaseFormatter``: The formatter associated to the data
                type.

        Raises:
        ------
            TypeError if a formatter cannot be found for this data
                type.

        Example usage:

        .. code-block:: pycon

            >>> from arctix import Summarizer
            >>> Summarizer.find_formatter(list)
            SequenceFormatter(max_items=5, num_spaces=2)
            >>> Summarizer.find_formatter(str)
            DefaultFormatter(max_characters=-1)
        """
        for object_type in data_type.__mro__:
            formatter = cls.registry.get(object_type, None)
            if formatter is not None:
                return formatter
        raise TypeError(f"Incorrect data type: {data_type}")

    @classmethod
    def load_state_dict(cls, state: dict) -> None:
        r"""Load the state values from a dict.

        Args:
        ----
            state_dict (dict): a dict with parameters

        Example usage:

        .. code-block:: pycon

            >>> from arctix import Summarizer
            >>> Summarizer.load_state_dict({object: {"max_characters": 10}})
            >>> summarizer = Summarizer()
            >>> summarizer.registry[object]
            DefaultFormatter(max_characters=10)
            >>> Summarizer.load_state_dict({object: {"max_characters": -1}})
            >>> summarizer.registry[object]
            DefaultFormatter(max_characters=-1)
        """
        for data_type, formatter in cls.registry.items():
            if (s := state.get(data_type)) is not None:
                formatter.load_state_dict(s)

    @classmethod
    def state_dict(cls) -> dict:
        r"""Return a dictionary containing state values.

        Returns:
        -------
            dict: the state values in a dict.

        Example usage:

        .. code-block:: pycon

            >>> from arctix import Summarizer
            >>> Summarizer.state_dict()  # doctest: +ELLIPSIS
            {<class 'collections.abc.Mapping'>: {'max_items': 5, 'num_spaces': 2},...
        """
        return {data_type: formatter.state_dict() for data_type, formatter in cls.registry.items()}

    @classmethod
    def set_max_characters(cls, max_characters: int) -> None:
        r"""Set the maximum of characters for the compatible formatter to
        the specified value.

        To be updated, the formatters need to implement the method
        ``set_max_characters``.

        Args:
        ----
            max_characters (int): Specifies the maximum of characters.

        Example usage:

        .. code-block:: pycon

            >>> from arctix import Summarizer
            >>> Summarizer.set_max_characters(10)
            >>> summarizer = Summarizer()
            >>> summarizer.registry[object]
            DefaultFormatter(max_characters=10)
            >>> Summarizer.set_max_characters(-1)
            >>> summarizer.registry[object]
            DefaultFormatter(max_characters=-1)
        """
        for formatter in cls.registry.values():
            if hasattr(formatter, "set_max_characters"):
                formatter.set_max_characters(max_characters)

    @classmethod
    def set_max_items(cls, max_items: int) -> None:
        r"""Set the maximum number of items for the compatible formatter
        to the specified value.

        To be updated, the formatters need to implement the method
        ``set_max_items``.

        Args:
        ----
            max_items (int): Specifies the maximum number of items to
                show.

        Example usage:

        .. code-block:: pycon

            >>> from arctix import Summarizer
            >>> Summarizer.set_max_items(10)
            >>> summarizer = Summarizer()
            >>> summarizer.registry[dict]
            MappingFormatter(max_items=10, num_spaces=2)
            >>> Summarizer.set_max_items(5)
            >>> summarizer.registry[dict]
            MappingFormatter(max_items=5, num_spaces=2)
        """
        for formatter in cls.registry.values():
            if hasattr(formatter, "set_max_items"):
                formatter.set_max_items(max_items)

    @classmethod
    def set_num_spaces(cls, num_spaces: int) -> None:
        r"""Set the maximum of items for the compatible formatter to the
        specified value.

        To be updated, the formatters need to implement the method
        ``set_num_spaces``.

        Args:
        ----
            num_spaces (int): Specifies the number of spaces for
                indentation.

        Example usage:

        .. code-block:: pycon

            >>> from arctix import Summarizer
            >>> Summarizer.set_num_spaces(4)
            >>> summarizer = Summarizer()
            >>> summarizer.registry[dict]
            MappingFormatter(max_items=5, num_spaces=4)
            >>> Summarizer.set_num_spaces(2)
            >>> summarizer.registry[dict]
            MappingFormatter(max_items=5, num_spaces=2)
        """
        for formatter in cls.registry.values():
            if hasattr(formatter, "set_num_spaces"):
                formatter.set_num_spaces(num_spaces)


def set_summarizer_options(
    max_characters: int | None = None, max_items: int | None = None, num_spaces: int | None = None
) -> None:
    r"""Set the ``Summarizer`` options.

    Note: It is recommended to use ``summarizer_options`` rather than
    this function.

    Args:
    ----
        max_characters (int or None, optional): Specifies the maximum
            number of characters to show. If ``None``, the maximum
            number of characters is unchanged. Default: ``None``
        max_items (int or None, optional): Specifies the maximum
            number of items to show. If ``None``, the maximum
            number of items is unchanged. Default: ``None``
        num_spaces (int or None, optional): Specifies the number of
            spaces for indentation. If ``None``, the number of  spaces
            for indentation is unchanged.  Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> from arctix import set_summarizer_options, summary
        >>> print(summary("abcdefghijklmnopqrstuvwxyz"))
        <class 'str'> abcdefghijklmnopqrstuvwxyz
        >>> set_summarizer_options(max_characters=10)
        >>> print(summary("abcdefghijklmnopqrstuvwxyz"))
        <class 'str'> abcdefghij...
        >>> set_summarizer_options(max_characters=-1)
        >>> print(summary("abcdefghijklmnopqrstuvwxyz"))
        <class 'str'> abcdefghijklmnopqrstuvwxyz
    """
    if max_characters is not None:
        Summarizer.set_max_characters(max_characters)
    if max_items is not None:
        Summarizer.set_max_items(max_items)
    if num_spaces is not None:
        Summarizer.set_num_spaces(num_spaces)


@contextmanager
def summarizer_options(**kwargs) -> None:
    r"""Context manager that temporarily changes the summarizer options.

    Accepted arguments are same as ``set_summarizer_options``.
    The context manager temporary change the configuration of
    ``Summarizer``. This context manager has no effect if
    ``Summarizer`` is not used.

    Args:
    ----
        **kwargs: Accepted arguments are same as
            ``set_summarizer_options``.

    Example usage:

    .. code-block:: pycon

        >>> from arctix import summarizer_options, summary
        >>> print(summary("abcdefghijklmnopqrstuvwxyz"))
        <class 'str'> abcdefghijklmnopqrstuvwxyz
        >>> with summarizer_options(max_characters=10):
        ...     print(summary("abcdefghijklmnopqrstuvwxyz"))
        ...
        <class 'str'> abcdefghij...
        >>> print(summary("abcdefghijklmnopqrstuvwxyz"))
        <class 'str'> abcdefghijklmnopqrstuvwxyz
    """
    state = Summarizer.state_dict()
    set_summarizer_options(**kwargs)
    try:
        yield
    finally:
        Summarizer.load_state_dict(state)



from __future__ import annotations

__all__ = [
    "BaseFormatter",
    "DefaultFormatter",
    "MappingFormatter",
    "SequenceFormatter",
    "SetFormatter",
]

from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from itertools import islice
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from arctix.utils.format import str_indent, str_mapping, str_sequence

if TYPE_CHECKING:
    from arctix.summarizer import BaseSummarizer

T = TypeVar("T")


class BaseFormatter(ABC, Generic[T]):
    r"""Define the base class to implement a formatter."""

    @abstractmethod
    def clone(self) -> BaseFormatter:
        r"""Return a copy of the formatter.

        Returns:
        -------
            ``BaseFormatter``: A copy of the formatter.

        Example usage:

        .. code-block:: pycon

            >>> from arctix.formatter import DefaultFormatter
            >>> formatter = DefaultFormatter()
            >>> formatter2 = formatter.clone()
            >>> formatter.set_max_characters(10)
            >>> formatter
            DefaultFormatter(max_characters=10)
            >>> formatter2
            DefaultFormatter(max_characters=-1)
        """

    @abstractmethod
    def equal(self, other: Any) -> bool:
        r"""Indicate if the other object is equal to the self object.

        Args:
        ----
            other: Specifies the other object to compare.

        Returns:
        -------
            bool: ``True`` if the objects are equal,
                otherwise ``False``.

        Example usage:

        .. code-block:: pycon

            >>> from arctix.formatter import DefaultFormatter
            >>> formatter = DefaultFormatter()
            >>> formatter.equal(DefaultFormatter())
            True
            >>> formatter.equal(DefaultFormatter(max_characters=10))
            False
        """

    @abstractmethod
    def format(self, summarizer: BaseSummarizer, value: T, depth: int, max_depth: int) -> str:
        r"""Format a value.

        Args:
        ----
            summarizer (``BaseSummarizer``): Specifies the summarizer.
            value: Specifies the value to summarize.

        Returns:
        -------
            str: The formatted value.

        Example usage:

        .. code-block:: pycon

            >>> from arctix import Summarizer
            >>> from arctix.formatter import DefaultFormatter
            >>> formatter = DefaultFormatter()
            >>> formatter.format(Summarizer(), 1)
            <class 'int'> 1
        """

    @abstractmethod
    def load_state_dict(self, state_dict: dict) -> None:
        r"""Load the state values from a dict.

        Args:
        ----
            state_dict (dict): a dict with parameters

        Example usage:

        .. code-block:: pycon

            >>> from arctix.formatter import DefaultFormatter
            >>> formatter = DefaultFormatter()
            >>> # Please take a look to the implementation of the state_dict
            >>> # function to know the expected structure
            >>> formatter.load_state_dict({"max_characters": 10})
            >>> formatter
            DefaultFormatter(max_characters=10)
        """

    @abstractmethod
    def state_dict(self) -> dict:
        r"""Return a dictionary containing state values.

        Example usage:
            dict: the state values in a dict.

        Example usage:

        .. code-block:: pycon

            >>> from arctix.formatter import DefaultFormatter
            >>> formatter = DefaultFormatter()
            >>> formatter.state_dict()
            {'max_characters': -1}
        """


class DefaultFormatter(BaseFormatter[Any]):
    r"""Implements the default formatter.

    Args:
    ----
        max_characters (int, optional): Specifies the maximum number
            of characters to show. If a negative value is provided,
            all the characters are shown. Default: ``-1``
    """

    def __init__(self, max_characters: int = -1) -> None:
        self.set_max_characters(max_characters)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(max_characters={self._max_characters:,})"

    def clone(self) -> DefaultFormatter:
        return self.__class__(max_characters=self._max_characters)

    def equal(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._max_characters == other._max_characters

    def format(
        self, summarizer: BaseSummarizer, value: Any, depth: int = 0, max_depth: int = 1
    ) -> str:
        if depth >= max_depth:
            return self._format(str(value))
        return f"{type(value)} {self._format(str(value))}"

    def _format(self, value: str) -> str:
        if self._max_characters >= 0 and len(value) > self._max_characters:
            value = value[: self._max_characters] + "..."
        return value

    def load_state_dict(self, state: dict) -> None:
        self._max_characters = state["max_characters"]

    def state_dict(self) -> dict:
        return {"max_characters": self._max_characters}

    def get_max_characters(self) -> int:
        r"""Gets the maximum number of characters to show.

        Returns:
        -------
            int: The maximum number of characters to show.

        Example usage:

        .. code-block:: pycon

            >>> from arctix.formatter import DefaultFormatter
            >>> formatter = DefaultFormatter()
            >>> formatter.get_max_characters()
            -1
        """
        return self._max_characters

    def set_max_characters(self, max_characters: int) -> None:
        r"""Set the maximum number of characters to show.

        Args:
        ----
            max_characters (int): Specifies the maximum number of
                characters to show.

        Raises:
        ------
            TypeError if ``max_characters`` is not an integer.

        Example usage:

        .. code-block:: pycon

            >>> from arctix.formatter import DefaultFormatter
            >>> formatter = DefaultFormatter()
            >>> formatter.set_max_characters(10)
            >>> formatter.get_max_characters()
            10
        """
        if not isinstance(max_characters, int):
            raise TypeError(
                "Incorrect type for max_characters. Expected int value but "
                f"received {max_characters}"
            )
        self._max_characters = max_characters


class BaseCollectionFormatter(BaseFormatter[T]):
    r"""Implement a base class to implement a formatter for
    ``Collection``.

    Args:
    ----
        max_items (int, optional): Specifies the maximum number
            of items to show. If a negative value is provided,
            all the items are shown. Default: ``5``
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. Default: ``2``.
    """

    def __init__(self, max_items: int = 5, num_spaces: int = 2) -> None:
        self.set_max_items(max_items)
        self.set_num_spaces(num_spaces)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(max_items={self._max_items:,}, "
            f"num_spaces={self._num_spaces})"
        )

    def clone(self) -> BaseCollectionFormatter:
        return self.__class__(max_items=self._max_items, num_spaces=self._num_spaces)

    def equal(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._max_items == other._max_items and self._num_spaces == other._num_spaces

    def load_state_dict(self, state: dict) -> None:
        self._max_items = state["max_items"]
        self._num_spaces = state["num_spaces"]

    def state_dict(self) -> dict:
        return {"max_items": self._max_items, "num_spaces": self._num_spaces}

    def get_max_items(self) -> int:
        r"""Gets the maximum number of items to show.

        Returns:
        -------
            int: The maximum number of items to show.

        Example usage:

        .. code-block:: pycon

            >>> from arctix.formatter import MappingFormatter
            >>> formatter = MappingFormatter()
            >>> formatter.get_max_items()
            5
        """
        return self._max_items

    def set_max_items(self, max_items: int) -> None:
        r"""Set the maximum number of items to show.

        Args:
        ----
            max_characters (int): Specifies the maximum number of
                items to show.

        Raises:
        ------
            TypeError if ``max_items`` is not an integer.

        Example usage:

        .. code-block:: pycon

            >>> from arctix.formatter import MappingFormatter
            >>> formatter = MappingFormatter()
            >>> formatter.set_max_items(10)
            >>> formatter.get_max_items()
            10
        """
        if not isinstance(max_items, int):
            raise TypeError(
                "Incorrect type for max_items. Expected int value but " f"received {max_items}"
            )
        self._max_items = max_items

    def get_num_spaces(self) -> int:
        r"""Gets the number of spaces for indentation.

        Returns:
        -------
            int: The number of spaces for indentation.

        Example usage:

        .. code-block:: pycon

            >>> from arctix.formatter import MappingFormatter
            >>> formatter = MappingFormatter()
            >>> formatter.get_num_spaces()
            2
        """
        return self._num_spaces

    def set_num_spaces(self, num_spaces: int) -> None:
        r"""Set the number of spaces for indentation.

        Args:
        ----
            max_characters (int): Specifies the number of spaces for
                indentation.

        Raises:
        ------
            TypeError if ``num_spaces`` is not an integer.
            TValueError if ``num_spaces`` is not a positive integer.

        Example usage:

        .. code-block:: pycon

            >>> from arctix.formatter import MappingFormatter
            >>> formatter = MappingFormatter()
            >>> formatter.set_num_spaces(4)
            >>> formatter.get_num_spaces()
            4
        """
        if not isinstance(num_spaces, int):
            raise TypeError(
                f"Incorrect type for num_spaces. Expected int value but received {num_spaces}"
            )
        if num_spaces < 0:
            raise ValueError(
                "Incorrect value for num_spaces. Expected a positive integer value but "
                f"received {num_spaces}"
            )
        self._num_spaces = num_spaces


class MappingFormatter(BaseCollectionFormatter[Mapping]):
    r"""Implements a formatter for ``Mapping``."""

    def format(
        self, summarizer: BaseSummarizer, value: Mapping, depth: int = 0, max_depth: int = 1
    ) -> str:
        if depth >= max_depth:
            return summarizer.summary(str(value), depth=depth + 1, max_depth=max_depth)
        typ = type(value)
        length = len(value)
        if length > 0:
            items = value.items()
            if self._max_items >= 0:
                items = islice(value.items(), self._max_items)
            value = str_mapping(
                {
                    key: summarizer.summary(val, depth=depth + 1, max_depth=max_depth)
                    for key, val in items
                },
                num_spaces=self._num_spaces,
            )
            if length > self._max_items and self._max_items >= 0:
                value = f"{value}\n..."
            value = f"(length={length:,})\n{value}"
        return str_indent(f"{typ} {value}", num_spaces=self._num_spaces)


class SequenceFormatter(BaseCollectionFormatter[Sequence]):
    r"""Implements a formatter for ``Sequence``."""

    def format(
        self, summarizer: BaseSummarizer, value: Sequence, depth: int = 0, max_depth: int = 1
    ) -> str:
        if depth >= max_depth:
            return summarizer.summary(str(value), depth=depth + 1, max_depth=max_depth)
        typ = type(value)
        length = len(value)
        if length > 0:
            if self._max_items >= 0:
                value = value[: self._max_items]
            value = str_sequence(
                [summarizer.summary(val, depth=depth + 1, max_depth=max_depth) for val in value],
                num_spaces=self._num_spaces,
            )
            if length > self._max_items and self._max_items > 0:
                value = f"{value}\n..."
            value = f"(length={length:,})\n{value}"
        return str_indent(f"{typ} {value}", num_spaces=self._num_spaces)


class SetFormatter(BaseCollectionFormatter[set]):
    r"""Implements a formatter for ``set``."""

    def format(
        self, summarizer: BaseSummarizer, value: set, depth: int = 0, max_depth: int = 1
    ) -> str:
        if depth >= max_depth:
            return summarizer.summary(str(value), depth=depth + 1, max_depth=max_depth)
        typ = type(value)
        length = len(value)
        if length > 0:
            if self._max_items >= 0:
                value = islice(value, self._max_items)
            value = str_sequence(
                [summarizer.summary(val, depth=depth + 1, max_depth=max_depth) for val in value],
                num_spaces=self._num_spaces,
            )
            if length > self._max_items and self._max_items > 0:
                value = f"{value}\n..."
            value = f"(length={length:,})\n{value}"
        return str_indent(f"{typ} {value}", num_spaces=self._num_spaces)



from __future__ import annotations

__all__ = ["str_indent", "str_mapping", "str_sequence"]

from collections.abc import Mapping, Sequence
from typing import Any


def str_indent(original: Any, num_spaces: int = 2) -> str:
    r"""Add indentations if the original string is a multi-lines string.

    Args:
    ----
        original: Specifies the original string. If the inputis not a
            string, it will be converted to a string with the function
            ``str``.
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. Default: ``2``.

    Returns:
    -------
        str: The indented string.

    Raises:
    ------
        RuntimeError if num_spaces is not a positive integer.

    Example usage:

    .. code-block:: pycon

        >>> from arctix.utils.format import str_indent
        >>> print(str_indent("string1\nstring2\n  string3", 4))
        string1
        string2
          string3
    """
    if num_spaces < 0:
        raise RuntimeError(
            f"Incorrect num_spaces. Expected a positive integer but received {num_spaces}"
        )
    formatted_str = str(original).split("\n")
    if len(formatted_str) == 1:  # single line
        return formatted_str[0]
    first = formatted_str.pop(0)
    formatted_str = "\n".join([(num_spaces * " ") + line for line in formatted_str])
    return first + "\n" + formatted_str


def str_mapping(mapping: Mapping, sorted_keys: bool = False, num_spaces: int = 2) -> str:
    r"""Computes a string representation of a mapping.

    This function was designed for flat dictionary. If you have a
    nested dictionary, you may consider other functions. Note that
    this function works for nested dict but the output may not be
    nice.

    Args:
    ----
        mapping (``Mapping``): Specifies the mapping.
        sorted_keys (bool, optional): Specifies if the key of the dict
            are sorted or not. Default: ``False``
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. Default: ``2``.

    Returns:
    -------
        str: The string representation of the mapping.

    Example usage:

    .. code-block:: pycon

        >>> from arctix.utils.format import str_mapping
        >>> print(str_mapping({"key1": "abc", "key2": "something\nelse"}))
        (key1): abc
        (key2): something
          else
    """
    lines = []
    for key, value in sorted(mapping.items()) if sorted_keys else mapping.items():
        lines.append(f"({key}): {str_indent(value, num_spaces=num_spaces)}")
    return "\n".join(lines)


def str_sequence(sequence: Sequence, num_spaces: int = 2) -> str:
    r"""Computes a string representation of a sequence.

    Args:
    ----
        sequence (``Sequence``): Specifies the sequence.
        num_spaces (int, optional): Specifies the number of spaces
            used for the indentation. Default: ``2``.

    Returns:
    -------
        str: The string representation of the sequence.

    Example usage:

    .. code-block:: pycon

        >>> from arctix.utils.format import str_sequence
        >>> print(str_sequence(["abc", "something\nelse"]))
        (0): abc
        (1): something
          else
    """
    lines = []
    for i, item in enumerate(sequence):
        lines.append(f"({i}): {str_indent(item, num_spaces=num_spaces)}")
    return "\n".join(lines)



from __future__ import annotations

from typing import Any
from unittest.mock import Mock

from arctix.formatter import BaseFormatter
from arctix.summarizer import BaseSummarizer, Summarizer
from arctix.utils.imports import check_numpy, is_numpy_available

if is_numpy_available():
    import numpy
else:
    numpy = Mock()  # pragma: no cover


class NDArrayFormatter(BaseFormatter[numpy.ndarray]):
    r"""Implement a formatter for ``numpy.ndarray``.

    Args:
    ----
        show_data (bool, optional): If ``True``, the returned string
            is the default string representation (``repr``).
            If ``False``, the returned string only contains the array
            metadata. Default: ``False``
    """

    def __init__(self, show_data: bool = False) -> None:
        check_numpy()
        self._show_data = bool(show_data)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(show_data={self._show_data})"

    def clone(self) -> NDArrayFormatter:
        return self.__class__(show_data=self._show_data)

    def equal(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._show_data == other._show_data

    def format(
        self, summarizer: BaseSummarizer, value: numpy.ndarray, depth: int = 0, max_depth: int = 1
    ) -> str:
        if self._show_data:
            return repr(value)
        return " | ".join(
            [
                f"{type(value)}",
                f"shape={value.shape}",
                f"dtype={value.dtype}",
            ]
        )

    def load_state_dict(self, state: dict) -> None:
        self._show_data = state["show_data"]

    def state_dict(self) -> dict:
        return {"show_data": self._show_data}

    def get_show_data(self) -> bool:
        r"""Indicates if the array data or metadata are shown.

        Returns:
        -------
            bool: ``True`` if the array data are shown,
                ``False`` if the array metadata are shown.

        Example usage:

        .. code-block:: pycon

            >>> from arctix._numpy import NDArrayFormatter
            >>> formatter = NDArrayFormatter()
            >>> formatter.get_show_data()
            False
        """
        return self._show_data

    def set_show_data(self, show_data: bool) -> None:
        r"""Set if the array data or metadata are shown.

        Args:
        ----
            show_data (bool): ``True`` if the array data are shown,
                ``False`` if the array metadata are shown.

        Raises:
        ------
            TypeError if ``show_data`` is not an boolean.

        Example usage:

        .. code-block:: pycon

            >>> from arctix._numpy import NDArrayFormatter
            >>> formatter = NDArrayFormatter()
            >>> formatter.set_show_data(True)
            >>> formatter.get_show_data()
            True
        """
        if not isinstance(show_data, bool):
            raise TypeError(
                "Incorrect type for show_data. Expected bool value but received {show_data}"
            )
        self._show_data = show_data


if is_numpy_available():  # pragma: no cover
    if not Summarizer.has_formatter(numpy.ndarray):
        Summarizer.add_formatter(numpy.ndarray, NDArrayFormatter())



from __future__ import annotations

__all__ = [
    "check_numpy",
    "check_torch",
    "is_numpy_available",
    "is_torch_available",
]

from importlib.util import find_spec


def is_numpy_available() -> bool:
    r"""Indicates if the NumPy package is installed or not."""
    return find_spec("numpy") is not None


def check_numpy() -> None:
    r"""Checks if the numpy package is installed.

    Raises:
        RuntimeError if the numpy package is not installed.
    """
    if not is_numpy_available():
        raise RuntimeError(
            "`numpy` package is required but not installed. "
            "You can install `numpy` package with the command:\n\n"
            "pip install numpy\n"
        )


def is_torch_available() -> bool:
    r"""Indicates if the torch package is installed or not."""
    return find_spec("torch") is not None


def check_torch() -> None:
    r"""Checks if the torch package is installed.

    Raises:
        RuntimeError if the torch package is not installed.
    """
    if not is_torch_available():
        raise RuntimeError(
            "`torch` package is required but not installed. "
            "You can install `torch` package with the command:\n\n"
            "pip install torch\n"
        )
